@@ -12,7 +12,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
- * $Id: pkcs11_lib.c 478 2011-01-22 17:33:07Z ludovic.rousseau $
+ * $Id$
  */
 
 #define __PKCS11_LIB_C__
@@ -57,7 +57,6 @@ int pkcs11_pass_login(pkcs11_handle_t *h, int nullok)
 
   /* check password length */
   if (!nullok && strlen(pin) == 0) {
-    free(pin);
     set_error("Empty passwords not allowed");
     return -1;
   }
@@ -65,7 +64,6 @@ int pkcs11_pass_login(pkcs11_handle_t *h, int nullok)
   /* perform pkcs #11 login */
   rv = pkcs11_login(h, pin);
   memset(pin, 0, strlen(pin));
-  free(pin);
   if (rv != 0) {
     set_error("pkcs11_login() failed: %s", get_error());
     return -1;
@@ -993,7 +991,7 @@ int load_pkcs11_module(const char *module, pkcs11_handle_t **hp)
   if (module_stat.st_mode & S_IWGRP || module_stat.st_mode & S_IWOTH
       || module_stat.st_uid != 0) {
     set_error("the pkcs #11 module MUST be owned by root and MUST NOT "
-              "be writeable by the group or others");
+              "be writable by the group or others");
     free(h);
     return -1;
   }
@@ -1026,15 +1024,66 @@ int load_pkcs11_module(const char *module, pkcs11_handle_t **hp)
 static int
 refresh_slots(pkcs11_handle_t *h)
 {
-  CK_ULONG i;
+  CK_ULONG i, slot_count;
+  CK_SLOT_ID_PTR slots;
+  CK_RV rv;
   int j;
 
+  slot_count = -1;
+  slots = NULL;
+  rv = h->fl->C_GetSlotList(FALSE, NULL, &slot_count);
+  if (rv != CKR_OK) {
+    set_error("C_GetSlotList() failed: 0x%08lX", rv);
+    return -1;
+  }
+
+  /* number of slots has changed */
+  if (slot_count != h->slot_count) {
+    free(h->slots);
+
+    /* get a list of all slots */
+	rv = h->fl->C_GetSlotList(FALSE, NULL, &h->slot_count);
+	if (rv != CKR_OK) {
+	  set_error("C_GetSlotList() failed: 0x%08lX", rv);
+	  return -1;
+	}
+	DBG1("number of slots (a): %ld", h->slot_count);
+	if (h->slot_count == 0) {
+	  set_error("there are no slots available");
+	  return -1;
+	}
+	slots = malloc(h->slot_count * sizeof(CK_SLOT_ID));
+	if (slots == NULL) {
+	  set_error("not enough free memory available");
+	  return -1;
+	}
+	h->slots = malloc(h->slot_count * sizeof(slot_t));
+	if (h->slots == NULL) {
+	  free(slots);
+	  set_error("not enough free memory available");
+	  return -1;
+	}
+	memset(h->slots, 0, h->slot_count * sizeof(slot_t));
+	rv = h->fl->C_GetSlotList(FALSE, slots, &h->slot_count);
+	if (rv != CKR_OK) {
+	  free(slots);
+	  set_error("C_GetSlotList() failed: 0x%08lX", rv);
+	  return -1;
+	}
+	DBG1("number of slots (b): %ld", h->slot_count);
+	/* show some information about the slots/tokens and setup slot info */
+	for (i = 0; i < h->slot_count; i++) {
+	  h->slots[i].id = slots[i];
+	}
+	free(slots);
+  }
+
   for (i = 0; i < h->slot_count; i++) {
     CK_SLOT_INFO sinfo;
     CK_TOKEN_INFO tinfo;
     CK_RV rv;
 
-    DBG1("slot %d:", i + 1);
+    DBG1("slot %ld:", i + 1);
     rv = h->fl->C_GetSlotInfo(h->slots[i].id, &sinfo);
     if (rv != CKR_OK) {
       set_error("C_GetSlotInfo() failed: 0x%08lX", rv);
@@ -1071,7 +1120,7 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
 {
   int rv;
   CK_ULONG i;
-  CK_SLOT_ID_PTR slots;
+  /* CK_SLOT_ID_PTR slots; */
   CK_INFO info;
   /*
    Set up arguments to allow native threads
@@ -1109,41 +1158,14 @@ int init_pkcs11_module(pkcs11_handle_t *h,int flag)
   DBG1("- flags: %04lx", info.flags);
   DBG1("- library description: %.32s", info.libraryDescription);
   DBG2("- library version: %hhd.%hhd", info.libraryVersion.major, info.libraryVersion.minor);
-  /* get a list of all slots */
-  rv = h->fl->C_GetSlotList(FALSE, NULL, &h->slot_count);
-  if (rv != CKR_OK) {
-    set_error("C_GetSlotList() failed: 0x%08lX", rv);
-    return -1;
-  }
-  DBG1("number of slots (a): %d", h->slot_count);
-  if (h->slot_count == 0) {
-    set_error("there are no slots available");
-    return -1;
-  }
-  slots = malloc(h->slot_count * sizeof(CK_SLOT_ID));
-  if (slots == NULL) {
-    set_error("not enough free memory available");
-    return -1;
-  }
-  h->slots = malloc(h->slot_count * sizeof(slot_t));
-  if (h->slots == NULL) {
-    free(slots);
-    set_error("not enough free memory available");
-    return -1;
-  }
-  memset(h->slots, 0, h->slot_count * sizeof(slot_t));
-  rv = h->fl->C_GetSlotList(FALSE, slots, &h->slot_count);
-  if (rv != CKR_OK) {
-    free(slots);
-    set_error("C_GetSlotList() failed: 0x%08lX", rv);
-    return -1;
-  }
-  DBG1("number of slots (b): %d", h->slot_count);
-  /* show some information about the slots/tokens and setup slot info */
-  for (i = 0; i < h->slot_count; i++) {
-    h->slots[i].id = slots[i];
-  }
-  free(slots);
+
+  /*
+   * As per PKCS#11 v2.2 we can call C_GetSlotList multiple times to check for
+   * added/removed slots
+   */
+  h->slot_count = -1;
+  h->slots = NULL;
+
   return refresh_slots(h);
 }
 
@@ -1385,7 +1407,7 @@ int pkcs11_login(pkcs11_handle_t *h, char *password)
 	  rv = h->fl->C_Login(h->session, CKU_USER, (unsigned char*)password, strlen(password));
   else
 	  rv = h->fl->C_Login(h->session, CKU_USER, NULL, 0);
-  if (rv != CKR_OK) {
+  if ((rv != CKR_OK) && (rv != CKR_USER_ALREADY_LOGGED_IN)) {
     set_error("C_Login() failed: 0x%08lX", rv);
     return -1;
   }
@@ -1516,7 +1538,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
         set_error("CertID length: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
     }
-    /* allocate enought space */
+    /* allocate enough space */
     id_value = malloc(cert_template[2].ulValueLen);
     if (id_value == NULL) {
         set_error("CertID malloc(%d): not enough free memory available", cert_template[2].ulValueLen);
@@ -1540,7 +1562,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
         set_error("Cert Length: C_GetAttributeValue() failed: 0x%08lX", rv);
         goto getlist_error;
       }
-    /* allocate enought space */
+    /* allocate enough space */
       cert_value = malloc(cert_template[3].ulValueLen);
       if (cert_value == NULL) {
         set_error("Cert Length malloc(%d): not enough free memory available", cert_template[3].ulValueLen);
@@ -1576,7 +1598,7 @@ cert_object_t **get_certificate_list(pkcs11_handle_t *h, int *ncerts)
     h->certs=certs;
     DBG1("Saving Certificate #%d:", h->cert_count + 1);
     certs[h->cert_count] = NULL;
-    DBG1("- type: %02x", cert_type);
+    DBG1("- type: %02lx", cert_type);
     DBG1("- id:   %02x", id_value[0]);
     h->certs[h->cert_count] = (cert_object_t *)calloc(sizeof(cert_object_t),1);
     if (h->certs[h->cert_count] == NULL) {
@@ -1628,14 +1650,11 @@ getlist_error:
 int get_private_key(pkcs11_handle_t *h, cert_object_t *cert) {
   CK_OBJECT_CLASS key_class = CKO_PRIVATE_KEY;
   CK_BBOOL key_sign = CK_TRUE;
-  CK_KEY_TYPE key_type = CKK_RSA; /* default, should be properly set */
   CK_ATTRIBUTE key_template[] = {
     {CKA_CLASS, &key_class, sizeof(key_class)}
     ,
     {CKA_SIGN, &key_sign, sizeof(key_sign)}
     ,
-    {CKA_KEY_TYPE, &key_type, sizeof(key_type)}
-    ,
     {CKA_ID, NULL, 0}
   };
   CK_OBJECT_HANDLE object;
@@ -1643,13 +1662,18 @@ int get_private_key(pkcs11_handle_t *h, cert_object_t *cert) {
   int rv;
 
   if (cert->private_key != CK_INVALID_HANDLE) {
-     /* we've alrady found the private key for this certificate */
+     /* we've already found the private key for this certificate */
      return 0;
   }
 
-  key_template[3].pValue = cert->id;
-  key_template[3].ulValueLen = cert->id_length;
-  rv = h->fl->C_FindObjectsInit(h->session, key_template, 2);
+  /* search for a specific ID is any */
+  if (cert->id && cert->id_length) {
+	  key_template[2].pValue = cert->id;
+	  key_template[2].ulValueLen = cert->id_length;
+	  rv = h->fl->C_FindObjectsInit(h->session, key_template, 3);
+  } else {
+	  rv = h->fl->C_FindObjectsInit(h->session, key_template, 2);
+  }
   if (rv != CKR_OK) {
     set_error("C_FindObjectsInit() failed: 0x%08lX", rv);
     return -1;
@@ -1720,7 +1744,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
   }
   /* compute hash-value */
   SHA1(data, length, &hash[15]);
-  DBG5("hash[%d] = [...:%02x:%02x:%02x:...:%02x]", sizeof(hash),
+  DBG5("hash[%ld] = [...:%02x:%02x:%02x:...:%02x]", sizeof(hash),
       hash[15], hash[16], hash[17], hash[sizeof(hash) - 1]);
   /* sign the token */
   rv = h->fl->C_SignInit(h->session, &mechanism, cert->private_key);
@@ -1742,7 +1766,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
       free(*signature);
       *signature = NULL;
       *signature_length *= 2;
-      DBG1("increased signature buffer-length to %d", *signature_length);
+      DBG1("increased signature buffer-length to %ld", *signature_length);
     } else if (rv != CKR_OK) {
       free(*signature);
       *signature = NULL;
@@ -1750,7 +1774,7 @@ int sign_value(pkcs11_handle_t *h, cert_object_t *cert, CK_BYTE *data,
       return -1;
     }
   }
-  DBG5("signature[%d] = [%02x:%02x:%02x:...:%02x]", *signature_length,
+  DBG5("signature[%ld] = [%02x:%02x:%02x:...:%02x]", *signature_length,
       (*signature)[0], (*signature)[1], (*signature)[2], (*signature)[*signature_length - 1]);
   return 0;
 }
