@@ -12,7 +12,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
- * $Id: pam_pkcs11.c 512 2012-04-07 16:55:20Z ludovic.rousseau $
+ * $Id$
  */
 
 /* We have to make this definitions before we include the pam header files! */
@@ -29,6 +29,10 @@
 #ifdef HAVE_SECURITY_PAM_EXT_H
 #include <security/pam_ext.h>
 #endif
+/* OpenPAM used on *BSD and OS X */
+#ifdef OPENPAM
+#include <security/openpam.h>
+#endif
 #include <syslog.h>
 #include <ctype.h>
 #include <string.h>
@@ -68,7 +72,7 @@ static int is_spaced_str(const char *str) {
 	return 1;
 }
 
-#ifndef HAVE_SECURITY_PAM_EXT_H
+#if !defined(HAVE_SECURITY_PAM_EXT_H) && !defined(OPENPAM)
 /*
  * implement pam utilities for older versions of pam.
  */
@@ -108,7 +112,9 @@ static int pam_prompt(pam_handle_t *pamh, int style, char **response, char *fmt,
   free(&resp[0]);
   return PAM_SUCCESS;
 }
+#endif
 
+#if !defined(HAVE_SECURITY_PAM_EXT_H) || defined(OPENPAM)
 static void
 pam_syslog(pam_handle_t *pamh, int priority, const char *fmt, ...)
 {
@@ -118,6 +124,27 @@ pam_syslog(pam_handle_t *pamh, int priority, const char *fmt, ...)
     vsyslog(priority, fmt, ap);
     va_end(ap);
 }
+
+/*
+ * With OpenPAM pam_prompt resp arg cannot be NULL, so this is just a wrapper.
+ */
+#undef pam_prompt
+#define pam_prompt(x, y, z, fmt, ...) pam_pkcs11_prompt((x), (y), (z), (fmt), ##__VA_ARGS__)
+
+static int pam_pkcs11_prompt(const pam_handle_t *pamh, int style, char **resp, const char *fmt, ...)
+{
+  char *response = NULL;
+  va_list va;
+  int ret = 0;
+
+  va_start(va, fmt);
+  ret = pam_vprompt(pamh, style, &response, fmt, va);
+  va_end(va);
+
+  free(response);
+
+  return ret;
+}
 #endif
 
 
@@ -199,7 +226,13 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
   char **issuer, **serial;
   const char *login_token_name = NULL;
 
-  pam_prompt(pamh, PAM_TEXT_INFO , NULL, _("Smartcard authentification starts"));
+#ifdef ENABLE_NLS
+  setlocale(LC_ALL, "");
+  bindtextdomain(PACKAGE, "/usr/share/locale");
+  textdomain(PACKAGE);
+#endif
+
+  pam_prompt(pamh, PAM_TEXT_INFO , NULL, _("Smartcard authentication starts"));
 
   /* first of all check whether debugging should be enabled */
   for (i = 0; i < argc; i++)
@@ -238,12 +271,6 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
 	  }
   }
 
-#ifdef ENABLE_NLS
-  setlocale(LC_ALL, "");
-  bindtextdomain(PACKAGE, "/usr/share/locale");
-  textdomain(PACKAGE);
-#endif
-
   /* init openssl */
   rv = crypto_init(&configuration->policy);
   if (rv != 0) {
@@ -580,7 +607,7 @@ PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc, cons
 			}
 			sleep(configuration->err_display_time);
 		}
-	goto auth_failed_nopw;
+        continue; /* try next certificate */
       } else if (rv != 1) {
         ERR1("verify_certificate() failed: %s", get_error());
         continue; /* try next certificate */
